# -*- coding: utf-8 -*-

# Copyright (c) 2002 - 2006 Detlev Offenbach <detlev@die-offenbachs.de>
#

"""
Module implementing the debug server.
"""

import sys
import os
import socket
import time
import signal
import subprocess

from PyQt4.QtCore import *
from PyQt4.QtNetwork import QTcpServer, QTcpSocket, QHostAddress

from KdeQt import KQMessageBox
from KdeQt.KQApplication import e4App

from DebugClients.Python.DebugProtocol import *
from Config import ClientDefaultCapabilities, ClientTypeAssociations
from BreakPointModel import BreakPointModel
from WatchPointModel import WatchPointModel
import Preferences
import Utilities

from eric4config import getConfig


class DebugServer(QTcpServer):
    """
    Class implementing the debug server embedded within the IDE.
    
    @signal clientOutput emitted after the client has sent some output
    @signal clientRawInputSent emitted after the data was sent to the debug client
    @signal clientLine(filename, lineno) emitted after the debug client has executed a
            line of code
    @signal clientStack(stack) emitted after the debug client has executed a
            line of code
    @signal clientVariables(variables) emitted after a variables dump has been received
    @signal clientVariable(variables) emitted after a dump for one class variable has
            been received
    @signal clientStatement(boolean) emitted after an interactive command has
            been executed. The parameter is 0 to indicate that the command is
            complete and 1 if it needs more input.
    @signal clientException(exception) emitted after an exception occured on the 
            client side
    @signal clientSyntaxError(exception) emitted after a syntax error has been detected
            on the client side
    @signal clientExit(exitcode) emitted after the client has exited
    @signal clientClearBreak(filename, lineno) emitted after the debug client
            has decided to clear a temporary breakpoint
    @signal clientBreakConditionError(fn, lineno) emitted after the client has signaled
            a syntax error in a breakpoint condition
    @signal clientClearWatch(condition) emitted after the debug client
            has decided to clear a temporary watchpoint
    @signal clientWatchConditionError(fn, lineno) emitted after the client has signaled
            a syntax error in a watchpoint condition
    @signal clientRawInput(prompt, echo) emitted after a raw input request was received
    @signal clientBanner(banner) emitted after the client banner was received
    @signal clientCapabilities(int capabilities, QString cltype) emitted after the clients
            capabilities were received
    @signal clientCompletionList(completionList, text) emitted after the client
            the commandline completion list and the reworked searchstring was
            received from the client
    @signal passiveDebugStarted emitted after the debug client has connected in
            passive debug mode
    @signal clientGone emitted if the client went away (planned or unplanned)
    @signal utPrepared(nrTests, exc_type, exc_value) emitted after the client has
            loaded a unittest suite
    @signal utFinished emitted after the client signalled the end of the unittest
    @signal utStartTest(testname, testdocu) emitted after the client has started 
            a test
    @signal utStopTest emitted after the client has finished a test
    @signal utTestFailed(testname, exc_info) emitted after the client reported 
            a failed test
    @signal utTestErrored(testname, exc_info) emitted after the client reported 
            an errored test
    @signal cyclopsError(fn, modfunc) emitted after the client reported an error
            while performing a Cyclops run
    """
    def __init__(self):
        """
        Constructor
        """
        QTcpServer.__init__(self)
        
        # the buffer size for the readLine call
        self.readLineBufferSize = 1024 * 1024
        
        # create our models
        self.breakpointModel = BreakPointModel(self)
        self.watchpointModel = WatchPointModel(self)
        self.watchSpecialCreated = \
            self.trUtf8("created", "must be same as in EditWatchpointDialog") 
        self.watchSpecialChanged = \
            self.trUtf8("changed", "must be same as in EditWatchpointDialog")
        
        if unicode(Preferences.getDebugger("NetworkInterface")) == "all":
            hostAddress = QHostAddress(QHostAddress.Any)
        else:
            hostAddress = QHostAddress(Preferences.getDebugger("NetworkInterface"))
        if Preferences.getDebugger("PassiveDbgEnabled"):
            socket = Preferences.getDebugger("PassiveDbgPort") # default: 42424
            self.listen(hostAddress, socket)
            self.passive = True
            self.passiveClientExited = False
            # set translation function
            if Preferences.getDebugger("PathTranslation"):
                self.translateRemote = \
                    unicode(Preferences.getDebugger("PathTranslationRemote"))
                self.translateLocal = \
                    unicode(Preferences.getDebugger("PathTranslationLocal"))
                self.translate = self.__remoteTranslation
            else:
                self.translate = self.__identityTranslation
        else:
            self.listen(hostAddress)
            self.passive = False
            # set translation function
            self.translate = self.__identityTranslation

        self.qsock = None
        self.debugging = False
        self.queue = []
        self.clientPID = 0
        self.clientType = Preferences.Prefs.settings.value(\
            'DebugClient/Type', QVariant('Python')).toString()
        self.lastClientType = ''
        # set default values for capabilities of clients
        self.clientCapabilities = {
            "Python" : ClientDefaultCapabilities["Python"],
            "Ruby"   : ClientDefaultCapabilities["Ruby"],
        }
        self.__autoClearShell = False
        
        self.connect(self, SIGNAL("clientCapabilities"), 
            self.__clientCapabilities)
        self.connect(self, SIGNAL("clientClearBreak"), self.__clientClearBreakPoint)
        self.connect(self, SIGNAL("clientClearWatch"), self.__clientClearWatchPoint)
        self.connect(self, SIGNAL("newConnection()"), self.__newConnection)
        
        self.connect(self.breakpointModel, 
            SIGNAL("rowsAboutToBeRemoved(const QModelIndex &, int, int)"), 
            self.__deleteBreakPoints)
        self.connect(self.breakpointModel,
            SIGNAL("dataAboutToBeChanged(const QModelIndex &, const QModelIndex &)"),
            self.__breakPointDataAboutToBeChanged)
        self.connect(self.breakpointModel,
            SIGNAL("dataChanged(const QModelIndex &, const QModelIndex &)"),
            self.__changeBreakPoints)
        self.connect(self.breakpointModel,
            SIGNAL("rowsInserted(const QModelIndex &, int, int)"),
            self.__addBreakPoints)
        
        self.connect(self.watchpointModel, 
            SIGNAL("rowsAboutToBeRemoved(const QModelIndex &, int, int)"), 
            self.__deleteWatchPoints)
        self.connect(self.watchpointModel,
            SIGNAL("dataAboutToBeChanged(const QModelIndex &, const QModelIndex &)"),
            self.__watchPointDataAboutToBeChanged)
        self.connect(self.watchpointModel, 
            SIGNAL("dataChanged(const QModelIndex &, const QModelIndex &)"),
            self.__changeWatchPoints)
        self.connect(self.watchpointModel,
            SIGNAL("rowsInserted(const QModelIndex &, int, int)"),
            self.__addWatchPoints)
        
        self.codec = QTextCodec.codecForName(sys.getdefaultencoding())

    def __identityTranslation(self, fn, remote2local=1):
        """
        Private method to perform the identity path translation.
        
        @param fn filename to be translated (string or QString)
        @param remote2local flag indicating the direction of translation
            (0 = local to remote, 1 = remote to local [default])
        @return translated filename (string)
        """
        return unicode(fn)
        
    def __remoteTranslation(self, fn, remote2local=1):
        """
        Private method to perform the path translation.
        
        @param fn filename to be translated (string or QString)
        @param remote2local flag indicating the direction of translation
            (0 = local to remote, 1 = remote to local [default])
        @return translated filename (string)
        """
        if remote2local:
            return unicode(fn).replace(self.translateRemote, self.translateLocal)
        else:
            return unicode(fn).replace(self.translateLocal, self.translateRemote)
        
    def startRemote(self, forProject = False):
        """
        Public method to start a remote interpreter.
        @param forProject flag indicating a project related action (boolean)
        """
        # only start the client, if we are not in passive mode
        if self.passive:
            return
            
        if self.clientPID:
            try:
                # get rid of dead child on Unix systems; if the child is
                # non cooperative, it will be killed
                status = os.waitpid(self.clientPID, os.WNOHANG)
                if status == (0, 0):
                    os.kill(self.clientPID, signal.SIGTERM)
                    time.sleep(1)
                    status = os.waitpid(self.clientPID, os.WNOHANG)
                    if status == (0, 0):
                        os.kill(self.clientPID, signal.SIGKILL)
            except:
                pass
            
        if forProject:
            self.__startRemoteForProject()
        else:
            if self.clientType == "Ruby":
                self.__startRemoteRuby()
            else:
                self.__startRemotePython()
        
    def __startRemotePython(self):
        """
        Private method to start a remote Python interpreter.
        """
        if Preferences.getDebugger("CustomPythonInterpreter"):
            interpreter = str(Preferences.getDebugger("PythonInterpreter"))
            if interpreter == "":
                interpreter = sys.executable
        else:
            interpreter = sys.executable
            
        debugClientType = unicode(Preferences.getDebugger("DebugClientType"))
        if debugClientType == "standard":
            debugClient = os.path.join(getConfig('ericDir'), 
                                       "DebugClients", "Python", "DebugClient.py")
        elif debugClientType == "threaded":
            debugClient = os.path.join(getConfig('ericDir'),
                                       "DebugClients", "Python", "DebugClientThreads.py")
        else:
            debugClient = str(Preferences.getDebugger("DebugClient"))
            if debugClient == "":
                debugClient = os.path.join(sys.path[0],
                                           "DebugClients", "Python", "DebugClient.py")
            
        redirect = str(Preferences.getDebugger("PythonRedirect"))
        noencoding = \
            Preferences.getDebugger("PythonNoEncoding") and '--no-encoding' or ''
        
        if Preferences.getDebugger("RemoteDbgEnabled"):
            ipaddr = socket.gethostbyname(socket.gethostname())
            rexec = str(Preferences.getDebugger("RemoteExecution"))
            rhost = str(Preferences.getDebugger("RemoteHost"))
            if rhost == "":
                rhost = "localhost"
            if rexec:
                args = Utilities.parseOptionString(rexec) + \
                       [rhost, interpreter, os.path.abspath(debugClient),
                        noencoding, str(self.serverPort()), redirect, ipaddr]
                args[0] = Utilities.getExecutablePath(args[0])
                try:
                    self.clientPID = subprocess.Popen(args).pid
                except OSError, msg:
                    KQMessageBox.critical(None,
                        self.trUtf8("Start Debugger"),
                        self.trUtf8(\
                            """<p>The debugger backend could not be started.<br />"""
                            """Reason: %1</p>""").arg(unicode(msg)))
                # set translation function
                if Preferences.getDebugger("PathTranslation"):
                    self.translateRemote = \
                        unicode(Preferences.getDebugger("PathTranslationRemote"))
                    self.translateLocal = \
                        unicode(Preferences.getDebugger("PathTranslationLocal"))
                    self.translate = self.__remoteTranslation
                else:
                    self.translate = self.__identityTranslation
                return
        
        # set translation function
        self.translate = self.__identityTranslation
        
        # setup the environment for the debugger
        if Preferences.getDebugger("DebugEnvironmentReplace"):
            clientEnv = {}
        else:
            clientEnv = os.environ.copy()
        envlist = Utilities.parseEnvironmentString(\
            Preferences.getDebugger("DebugEnvironment"))
        for el in envlist:
            try:
                key, value = el.split('=', 1)
                if value.startswith('"') or value.startswith("'"):
                    value = value[1:-1]
                clientEnv[unicode(key)] = unicode(value)
            except UnpackError:
                pass
        
        if Preferences.getDebugger("ConsoleDbgEnabled"):
            ccmd = unicode(Preferences.getDebugger("ConsoleDbgCommand"))
            if ccmd:
                args = Utilities.parseOptionString(ccmd) + \
                       [interpreter, os.path.abspath(debugClient),
                        noencoding, str(self.serverPort()), '0']
                args[0] = Utilities.getExecutablePath(args[0])
                try:
                    self.clientPID = subprocess.Popen(args, env = clientEnv).pid
                except OSError, msg:
                    KQMessageBox.critical(None,
                        self.trUtf8("Start Debugger"),
                        self.trUtf8(\
                            """<p>The debugger backend could not be started.<br />"""
                            """Reason: %1</p>""").arg(unicode(msg)))
                return
        
        try:
            self.clientPID = subprocess.Popen([interpreter, debugClient, 
                 noencoding, str(self.serverPort()), redirect],
                env = clientEnv).pid
        except OSError, msg:
            KQMessageBox.critical(None,
                self.trUtf8("Start Debugger"),
                self.trUtf8("""<p>The debugger backend could not be started.<br />"""
                    """Reason: %1</p>""").arg(unicode(msg)))

    def __startRemoteRuby(self):
        """
        Private method to start a remote Ruby interpreter.
        """
        interpreter = str(Preferences.getDebugger("RubyInterpreter"))
        if interpreter == "":
            interpreter = "/usr/bin/ruby"
        debugClient = os.path.join(getConfig('ericDir'), 
            "DebugClients", "Ruby", "DebugClient.rb")
        
        redirect = str(Preferences.getDebugger("RubyRedirect"))
        
        if Preferences.getDebugger("RemoteDbgEnabled"):
            ipaddr = socket.gethostbyname(socket.gethostname())
            rexec = str(Preferences.getDebugger("RemoteExecution"))
            rhost = str(Preferences.getDebugger("RemoteHost"))
            if rhost == "":
                rhost = "localhost"
            if rexec:
                args = Utilities.parseOptionString(rexec) + \
                       [rhost, interpreter, os.path.abspath(debugClient),
                        str(self.serverPort()), redirect, ipaddr]
                args[0] = Utilities.getExecutablePath(args[0])
                try:
                    self.clientPID = subprocess.Popen(args).pid
                except OSError, msg:
                    KQMessageBox.critical(None,
                        self.trUtf8("Start Debugger"),
                        self.trUtf8(\
                            """<p>The debugger backend could not be started.<br />"""
                            """Reason: %1</p>""").arg(unicode(msg)))
                # set translation function
                if Preferences.getDebugger("PathTranslation"):
                    self.translateRemote = \
                        unicode(Preferences.getDebugger("PathTranslationRemote"))
                    self.translateLocal = \
                        unicode(Preferences.getDebugger("PathTranslationLocal"))
                    self.translate = self.__remoteTranslation
                else:
                    self.translate = self.__identityTranslation
                return
        
        # set translation function
        self.translate = self.__identityTranslation
        
        # setup the environment for the debugger
        if Preferences.getDebugger("DebugEnvironmentReplace"):
            clientEnv = {}
        else:
            clientEnv = os.environ.copy()
        envlist = Utilities.parseEnvironmentString(\
            Preferences.getDebugger("DebugEnvironment"))
        for el in envlist:
            try:
                key, value = el.split('=', 1)
                if value.startswith('"') or value.startswith("'"):
                    value = value[1:-1]
                clientEnv[unicode(key)] = unicode(value)
            except UnpackError:
                pass
        
        if Preferences.getDebugger("ConsoleDbgEnabled"):
            ccmd = unicode(Preferences.getDebugger("ConsoleDbgCommand"))
            if ccmd:
                args = Utilities.parseOptionString(ccmd) + \
                       [interpreter, os.path.abspath(debugClient),
                        str(self.serverPort()), '0']
                args[0] = Utilities.getExecutablePath(args[0])
                try:
                    self.clientPID = subprocess.Popen(args, env = clientEnv).pid
                except OSError, msg:
                    KQMessageBox.critical(None,
                        self.trUtf8("Start Debugger"),
                        self.trUtf8(\
                            """<p>The debugger backend could not be started.<br />"""
                            """Reason: %1</p>""").arg(unicode(msg)))
                return
        
        try:
            self.clientPID = subprocess.Popen([interpreter, debugClient, 
                 str(self.serverPort()), redirect],
                env = clientEnv).pid
        except OSError, msg:
            KQMessageBox.critical(None,
                self.trUtf8("Start Debugger"),
                self.trUtf8("""<p>The debugger backend could not be started.<br />"""
                    """Reason: %1</p>""").arg(unicode(msg)))

    def __startRemoteForProject(self):
        """
        Private method to start a remote interpreter for a project.
        """
        project = e4App().getObject("Project")
        if not project.isDebugPropertiesLoaded():
            if self.clientType == "Ruby":
                self.__startRemoteRuby()
            else:
                self.__startRemotePython()
            return
        
        # start debugger with project specific settings
        interpreter = project.getDebugProperty("INTERPRETER")
        debugClient = project.getDebugProperty("DEBUGCLIENT")
        
        redirect = str(project.getDebugProperty("REDIRECT"))
        noencoding = \
            project.getDebugProperty("NOENCODING") and '--no-encoding' or ''
        
        if project.getDebugProperty("REMOTEDEBUGGER"):
            ipaddr = socket.gethostbyname(socket.gethostname())
            rexec = project.getDebugProperty("REMOTECOMMAND")
            rhost = project.getDebugProperty("REMOTEHOST")
            if rhost == "":
                rhost = "localhost"
            if rexec:
                args = Utilities.parseOptionString(rexec) + \
                       [rhost, interpreter, os.path.abspath(debugClient),
                        noencoding, str(self.serverPort()), redirect, ipaddr]
                args[0] = Utilities.getExecutablePath(args[0])
                try:
                    self.clientPID = subprocess.Popen(args).pid
                except OSError, msg:
                    KQMessageBox.critical(None,
                        self.trUtf8("Start Debugger"),
                        self.trUtf8(\
                            """<p>The debugger backend could not be started.<br />"""
                            """Reason: %1</p>""").arg(unicode(msg)))
                # set translation function
                if project.getDebugProperty("PATHTRANSLATION"):
                    self.translateRemote = project.getDebugProperty("REMOTEPATH")
                    self.translateLocal = project.getDebugProperty("LOCALPATH")
                    self.translate = self.__remoteTranslation
                else:
                    self.translate = self.__identityTranslation
                return
        
        # set translation function
        self.translate = self.__identityTranslation
        
        # setup the environment for the debugger
        if project.getDebugProperty("ENVIRONMENTOVERRIDE"):
            clientEnv = {}
        else:
            clientEnv = os.environ.copy()
        envlist = Utilities.parseEnvironmentString(\
            project.getDebugProperty("ENVIRONMENTSTRING"))
        for el in envlist:
            try:
                key, value = el.split('=', 1)
                if value.startswith('"') or value.startswith("'"):
                    value = value[1:-1]
                clientEnv[unicode(key)] = unicode(value)
            except UnpackError:
                pass
        
        if project.getDebugProperty("CONSOLEDEBUGGER"):
            ccmd = unicode(project.getDebugProperty("CONSOLECOMMAND"))
            if ccmd:
                args = Utilities.parseOptionString(ccmd) + \
                       [interpreter, os.path.abspath(debugClient),
                        noencoding, str(self.serverPort()), '0']
                args[0] = Utilities.getExecutablePath(args[0])
                try:
                    self.clientPID = subprocess.Popen(args, env = clientEnv).pid
                except OSError, msg:
                    KQMessageBox.critical(None,
                        self.trUtf8("Start Debugger"),
                        self.trUtf8(\
                            """<p>The debugger backend could not be started.<br />"""
                            """Reason: %1</p>""").arg(unicode(msg)))
                return
        
        try:
            self.clientPID = subprocess.Popen([interpreter, debugClient, 
                 noencoding, str(self.serverPort()), redirect],
                env = clientEnv).pid
        except OSError, msg:
            KQMessageBox.critical(None,
                self.trUtf8("Start Debugger"),
                self.trUtf8("""<p>The debugger backend could not be started.<br />"""
                    """Reason: %1</p>""").arg(unicode(msg)))

    def __clientClearBreakPoint(self, fn, lineno):
        """
        Private slot to handle the clientClearBreak signal.
        
        @param fn filename of breakpoint to clear (string or QString)
        @param lineno line number of breakpoint to clear (integer)
        """
        if self.debugging:
            index = self.breakpointModel.getBreakPointIndex(fn, lineno)
            self.breakpointModel.deleteBreakPointByIndex(index)

    def __deleteBreakPoints(self, parentIndex, start, end):
        """
        Private slot to delete breakpoints.
        
        @param parentIndex index of parent item (QModelIndex)
        @param start start row (integer)
        @param end end row (integer)
        """
        if self.debugging:
            for row in range(start, end+1):
                index = self.breakpointModel.index(row, 0, parentIndex)
                fn, lineno = self.breakpointModel.getBreakPointByIndex(index)[0:2]
                self.__remoteBreakpoint(fn, lineno, False)

    def __changeBreakPoints(self, startIndex, endIndex):
        """
        Private slot to set changed breakpoints.
        
        @param indexes indexes of changed breakpoints.
        """
        if self.debugging:
            self.__addBreakPoints(QModelIndex(), startIndex.row(), endIndex.row())

    def __breakPointDataAboutToBeChanged(self, startIndex, endIndex):
        """
        Private slot to handle the dataAboutToBeChanged signal of the breakpoint model.
        
        @param startIndex start index of the rows to be changed (QModelIndex)
        @param endIndex end index of the rows to be changed (QModelIndex)
        """
        if self.debugging:
            self.__deleteBreakPoints(QModelIndex(), startIndex.row(), endIndex.row())
        
    def __addBreakPoints(self, parentIndex, start, end):
        """
        Private slot to add breakpoints.
        
        @param parentIndex index of parent item (QModelIndex)
        @param start start row (integer)
        @param end end row (integer)
        """
        if self.debugging:
            for row in range(start, end+1):
                index = self.breakpointModel.index(row, 0, parentIndex)
                fn, line, cond, temp, enabled, ignorecount = \
                    self.breakpointModel.getBreakPointByIndex(index)[:6]
                self.__remoteBreakpoint(fn, line, True, cond, temp)
                if not enabled:
                    self.__remoteBreakpointEnable(fn, line, False)
                if ignorecount:
                    self.__remoteBreakpointIgnore(fn, line, ignorecount)

    def __makeWatchCondition(self, cond, special):
        """
        Private method to construct the condition string.
        
        @param cond condition (string or QString)
        @param special special condition (string or QString)
        @return condition string (QString)
        """
        special = unicode(special)
        if special == "":
            _cond = unicode(cond)
        else:
            if special == unicode(self.watchSpecialCreated):
                _cond = "%s ??created??" % cond
            elif special == unicode(self.watchSpecialChanged):
                _cond = "%s ??changed??" % cond
        return _cond
        
    def __splitWatchCondition(self, cond):
        """
        Private method to split a remote watchpoint condition.
        
        @param cond remote condition (string or QString)
        @return tuple of local condition (string) and special condition (string)
        """
        cond = unicode(cond)
        if cond.endswith(" ??created??"):
            cond, special = cond.split()
            special = unicode(self.watchSpecialCreated)
        elif cond.endswith(" ??changed??"):
            cond, special = cond.split()
            special = unicode(self.watchSpecialChanged)
        else:
            return cond, ""
        
    def __clientClearWatchPoint(self, condition):
        """
        Private slot to handle the clientClearWatch signal.
        
        @param condition condition of watchpoint to clear (string or QString)
        """
        if self.debugging:
            cond, special = self.__splitWatchCondition(condition)
            index = self.watchpointModel.getWatchPointIndex(cond, special)
            self.watchpoint.deleteWatchPointByIndex(index)
        
    def __deleteWatchPoints(self, parentIndex, start, end):
        """
        Private slot to delete watchpoints.
        
        @param parentIndex index of parent item (QModelIndex)
        @param start start row (integer)
        @param end end row (integer)
        """
        if self.debugging:
            for row in range(start, end+1):
                index = self.watchpointModel.index(row, 0, parentIndex)
                cond, special = self.watchpointModel.getWatchPointByIndex(index)[0:2]
                cond = self.__makeWatchCondition(cond, special)
                self.__remoteWatchpoint(cond, False)
        
    def __watchPointDataAboutToBeChanged(self, startIndex, endIndex):
        """
        Private slot to handle the dataAboutToBeChanged signal of the watchpoint model.
        
        @param startIndex start index of the rows to be changed (QModelIndex)
        @param endIndex end index of the rows to be changed (QModelIndex)
        """
        if self.debugging:
            self.__deleteWatchPoints(QModelIndex(), startIndex.row(), endIndex.row())
        
    def __addWatchPoints(self, parentIndex, start, end):
        """
        Private slot to set a watchpoint.
        
        @param parentIndex index of parent item (QModelIndex)
        @param start start row (integer)
        @param end end row (integer)
        """
        if self.debugging:
            for row in range(start, end+1):
                index = self.watchpointModel.index(row, 0, parentIndex)
                cond, special, temp, enabled, ignorecount = \
                    self.watchpointModel.getWatchPointByIndex(index)[:5]
                cond = self.__makeWatchCondition(cond, special)
                self.__remoteWatchpoint(cond, True, temp)
                if not enabled:
                    self.__remoteWatchpointEnable(cond, False)
                if ignorecount:
                    self.__remoteWatchpointIgnore(cond, ignorecount)
        
    def __changeWatchPoints(self, startIndex, endIndex):
        """
        Private slot to set changed watchpoints.
        
        @param startIndex start index of the rows to be changed (QModelIndex)
        @param endIndex end index of the rows to be changed (QModelIndex)
        """
        if self.debugging:
            self.__addWatchPoints(QModelIndex(), startIndex.row(), endIndex.row())
        
    def __clientCapabilities(self, cap, clType):
        """
        Private slot to handle the reporting of the clients capabilities.
        
        @param cap client capabilities (integer)
        @param clType type of the debug client (QString)
        """
        self.clientCapabilities[str(clType)] = cap
        
    def getClientCapabilities(self, type):
        """
        Public method to retrieve the debug clients capabilities.
        
        @param type debug client type (string)
        @return debug client capabilities (integer)
        """
        try:
            return self.clientCapabilities[type]
        except KeyError:
            return 0    # no capabilities
        
    def __newConnection(self):
        """
        Private slot to handle a new connection.
        """
        sock = self.nextPendingConnection()
        peerAddress = sock.peerAddress().toString()
        if Preferences.getDebugger("AllowedHosts").indexOf(peerAddress) == -1:
            # the peer is not allowed to connect
            KQMessageBox.warning(None,
                self.trUtf8("Connection from illegal host"),
                self.trUtf8("""<p>A connection was attempted by the"""
                    """ illegal host <b>%1</b>. Connection refused.</p>""")\
                    .arg(peerAddress))
            sock.abort()
            return

        # If we already have a connection, refuse this one.  It will be closed
        # automatically.
        if self.qsock is not None:
            return

        self.connect(sock,SIGNAL('readyRead()'),self.__parseClientLine)
        self.connect(sock,SIGNAL('connectionClosed()'),self.startClient)

        self.qsock = sock

        # Get the remote clients capabilities
        self.remoteCapabilities()
        
        # Perform actions neccessary, if client type has changed
        if self.lastClientType != self.clientType:
            self.lastClientType = self.clientType
            self.remoteBanner()
        elif self.__autoClearShell:
            self.__autoClearShell = False
            self.remoteBanner()
        
        # Send commands that were waiting for the connection.
        for cmd in self.queue:
            self.qsock.write(str(cmd))

        self.queue = []

    def shutdownServer(self):
        """
        Public method to cleanly shut down.
        
        It closes our socket and shuts down
        the debug client. (Needed on Win OS)
        """
        if self.qsock is None:
            return
            
        # do not want any slots called during shutdown
        self.disconnect(self.qsock, SIGNAL('connectionClosed()'), self.startClient)
        self.disconnect(self.qsock, SIGNAL('readyRead()'), self.__parseClientLine)
        
        # close down socket, and shut down client as well.
        self.__sendCommand('%s\n' % RequestShutdown)
        self.qsock.flush()
        
        self.qsock.close()
        
        # reinitialize
        self.qsock = None
        self.queue = []

    def remoteEnvironment(self, env):
        """
        Public method to set the environment for a program to debug, run, ...
        
        @param env environment settings (string)
        """
        envlist = Utilities.parseEnvironmentString(env)
        envdict = {}
        for el in envlist:
            try:
                key, value = el.split('=', 1)
                if value.startswith('"') or value.startswith("'"):
                    value = value[1:-1]
                envdict[unicode(key)] = unicode(value)
            except UnpackError:
                pass
        self.__sendCommand('%s%s\n' % (RequestEnv, str(envdict)))
        
    def remoteLoad(self, fn, argv, wd, env, autoClearShell = True,
                   tracePython = False, autoContinue = True, forProject = False):
        """
        Public method to load a new program to debug.
        
        @param fn the filename to debug (string)
        @param argv the commandline arguments to pass to the program (string or QString)
        @param wd the working directory for the program (string)
        @param env environment settings (string)
        @keyparam autoClearShell flag indicating, that the interpreter window should
            be cleared (boolean)
        @keyparam tracePython flag indicating if the Python library should be traced
            as well (boolean)
        @keyparam autoContinue flag indicating, that the debugger should not stop
            at the first executable line (boolean)
        @keyparam forProject flag indicating a project related action (boolean)
        """
        self.__autoClearShell = autoClearShell
        self.__autoContinue = autoContinue
        
        # Restart the client
        try:
            self.__setClientType(ClientTypeAssociations[os.path.splitext(fn)[1]])
        except KeyError:
            self.__setClientType('Python')    # assume it is a Python file
        self.startClient(False, forProject=forProject)
        
        self.remoteEnvironment(env)
        
        wd = self.translate(wd, 0)
        fn = self.translate(os.path.abspath(unicode(fn)), 0)
        self.__sendCommand('%s%s|%s|%s|%d\n' % \
            (RequestLoad, unicode(wd),
             unicode(fn), str(Utilities.parseOptionString(argv)),tracePython))
        self.debugging = True
        self.__restoreBreakpoints()
        self.__restoreWatchpoints()

    def remoteRun(self, fn, argv, wd, env, autoClearShell = True,
                  forProject = False):
        """
        Public method to load a new program to run.
        
        @param fn the filename to run (string)
        @param argv the commandline arguments to pass to the program (string or QString)
        @param wd the working directory for the program (string)
        @param env environment settings (string)
        @keyparam autoClearShell flag indicating, that the interpreter window should
            be cleared (boolean)
        @keyparam forProject flag indicating a project related action (boolean)
        """
        self.__autoClearShell = autoClearShell
        
        # Restart the client
        try:
            self.__setClientType(ClientTypeAssociations[os.path.splitext(fn)[1]])
        except KeyError:
            self.__setClientType('Python')    # assume it is a Python file
        self.startClient(False, forProject=forProject)
        
        self.remoteEnvironment(env)
        
        wd = self.translate(wd, 0)
        fn = self.translate(os.path.abspath(unicode(fn)), 0)
        self.__sendCommand('%s%s|%s|%s\n' % \
            (RequestRun, unicode(wd),
             unicode(fn), str(Utilities.parseOptionString(argv))))
        self.debugging = False

    def remoteCoverage(self, fn, argv, wd, env, autoClearShell = True,
                       erase = False, exclude = "", forProject = False):
        """
        Public method to load a new program to collect coverage data.
        
        @param fn the filename to run (string)
        @param argv the commandline arguments to pass to the program (string or QString)
        @param wd the working directory for the program (string)
        @param env environment settings (string)
        @keyparam autoClearShell flag indicating, that the interpreter window should
            be cleared (boolean)
        @keyparam erase flag indicating that coverage info should be 
            cleared first (boolean)
        @keyparam exclude searchpattern marking line to exclude from coverage
        @keyparam forProject flag indicating a project related action (boolean)
        """
        self.__autoClearShell = autoClearShell
        
        # Restart the client
        try:
            self.__setClientType(ClientTypeAssociations[os.path.splitext(fn)[1]])
        except KeyError:
            self.__setClientType('Python')    # assume it is a Python file
        self.startClient(False, forProject=forProject)
        
        self.remoteEnvironment(env)
        
        wd = self.translate(wd, 0)
        fn = self.translate(os.path.abspath(unicode(fn)), 0)
        self.__sendCommand('%s%s@@%s@@%s@@%d@@%s\n' % \
            (RequestCoverage, unicode(wd),
             unicode(fn),str(Utilities.parseOptionString(argv)),erase,unicode(exclude)))
        self.debugging = False

    def remoteProfile(self, fn, argv, wd, env, autoClearShell = True,
                      erase = False, forProject = False):
        """
        Public method to load a new program to collect profiling data.
        
        @param fn the filename to run (string)
        @param argv the commandline arguments to pass to the program (string or QString)
        @param wd the working directory for the program (string)
        @param env environment settings (string)
        @keyparam autoClearShell flag indicating, that the interpreter window should
            be cleared (boolean)
        @keyparam erase flag indicating that timing info should be cleared first (boolean)
        @keyparam forProject flag indicating a project related action (boolean)
        """
        self.__autoClearShell = autoClearShell
        
        # Restart the client
        try:
            self.__setClientType(ClientTypeAssociations[os.path.splitext(fn)[1]])
        except KeyError:
            self.__setClientType('Python')    # assume it is a Python file
        self.startClient(False, forProject=forProject)
        
        self.remoteEnvironment(env)
        
        wd = self.translate(wd, 0)
        fn = self.translate(os.path.abspath(unicode(fn)), 0)
        self.__sendCommand('%s%s|%s|%s|%d\n' % \
            (RequestProfile, unicode(wd),
             unicode(fn),str(Utilities.parseOptionString(argv)),erase))
        self.debugging = False

    def remoteCyclops(self, fn, argv, wd, env, modfunc, reports,
                      autoClearShell = True, forProject = False):
        """
        Public method to load a new program to collect profiling data.
        
        @param fn the filename to run (string)
        @param argv the commandline arguments to pass to the program (string or QString)
        @param wd the working directory for the program (string)
        @param env environment settings (string)
        @param modfunc name of a module function which is the main 
            entry point (string)
        @param reports bit mask specifying the reports wanted (integer)
        @keyparam autoClearShell flag indicating, that the interpreter window should
            be cleared (boolean)
        @keyparam forProject flag indicating a project related action (boolean)
        """
        self.__autoClearShell = autoClearShell
        
        # Restart the client
        try:
            self.__setClientType(ClientTypeAssociations[os.path.splitext(fn)[1]])
        except KeyError:
            self.__setClientType('Python')    # assume it is a Python file
        self.startClient(False, forProject=forProject)
        
        self.remoteEnvironment(env)
        
        wd = self.translate(wd, 0)
        fn = self.translate(os.path.abspath(unicode(fn)), 0)
        self.__sendCommand('%s%s|%s|%s|%s|%d\n' % \
            (RequestCyclops, unicode(wd),
             unicode(fn),str(Utilities.parseOptionString(argv)),str(modfunc),reports))
        self.debugging = False

    def remoteStatement(self,stmt):
        """
        Public method to execute a Python statement.  
        
        @param stmt the Python statement to execute (string). It
              should not have a trailing newline.
        """
        self.__sendCommand('%s\n' % stmt)
        self.__sendCommand(RequestOK + '\n')

    def remoteStep(self):
        """
        Public method to single step the debugged program.
        """
        self.__sendCommand(RequestStep + '\n')

    def remoteStepOver(self):
        """
        Public method to step over the debugged program.
        """
        self.__sendCommand(RequestStepOver + '\n')

    def remoteStepOut(self):
        """
        Public method to step out the debugged program.
        """
        self.__sendCommand(RequestStepOut + '\n')

    def remoteStepQuit(self):
        """
        Public method to stop the debugged program.
        """
        self.__sendCommand(RequestStepQuit + '\n')

    def remoteContinue(self, special=0):
        """
        Public method to continue the debugged program.
        
        @param special flag indicating a special continue operation
        """
        self.__sendCommand('%s%d\n' % (RequestContinue, special))

    def __remoteBreakpoint(self,fn,line,set,cond=None,temp=0):
        """
        Private method to set or clear a breakpoint.
        
        @param fn filename the breakpoint belongs to (string)
        @param line linenumber of the breakpoint (int)
        @param set flag indicating setting or resetting a breakpoint (boolean)
        @param cond condition of the breakpoint (string)
        @param temp flag indicating a temporary breakpoint (boolean)
        """
        fn = self.translate(fn, 0)
        self.__sendCommand('%s%s,%d,%d,%d,%s\n' % (RequestBreak,fn,line,temp,set,cond))
        
    def __remoteBreakpointEnable(self,fn,line,enable):
        """
        Private method to enable or disable a breakpoint.
        
        @param fn filename the breakpoint belongs to (string)
        @param line linenumber of the breakpoint (int)
        @param enable flag indicating enabling or disabling a breakpoint (boolean)
        """
        fn = self.translate(fn, 0)
        self.__sendCommand('%s%s,%d,%d\n' % (RequestBreakEnable,fn,line,enable))
        
    def __remoteBreakpointIgnore(self,fn,line,count):
        """
        Private method to ignore a breakpoint the next couple of occurences.
        
        @param fn filename the breakpoint belongs to (string)
        @param line linenumber of the breakpoint (int)
        @param count number of occurences to ignore (int)
        """
        fn = self.translate(fn, 0)
        self.__sendCommand('%s%s,%d,%d\n' % (RequestBreakIgnore,fn,line,count))
        
    def __remoteWatchpoint(self,cond,set,temp=0):
        """
        Private method to set or clear a watchpoint.
        
        @param cond condition of the watchpoint (string)
        @param set flag indicating setting or resetting a watchpoint (boolean)
        @param temp flag indicating a temporary watchpoint (boolean)
        """
        # cond is combination of cond and special (s. old watchpoint viewer)
        self.__sendCommand('%s%s,%d,%d\n' % (RequestWatch,cond,temp,set))
    
    def __remoteWatchpointEnable(self,cond,enable):
        """
        Private method to enable or disable a watchpoint.
        
        @param cond condition of the watchpoint (string)
        @param enable flag indicating enabling or disabling a watchpoint (boolean)
        """
        # cond is combination of cond and special (s. old watchpoint viewer)
        self.__sendCommand('%s%s,%d\n' % (RequestWatchEnable,cond,enable))
    
    def __remoteWatchpointIgnore(self,cond,count):
        """
        Private method to ignore a watchpoint the next couple of occurences.
        
        @param cond condition of the watchpoint (string)
        @param count number of occurences to ignore (int)
        """
        # cond is combination of cond and special (s. old watchpoint viewer)
        self.__sendCommand('%s%s,%d\n' % (RequestWatchIgnore,cond,count))
    
    def remoteRawInput(self,s):
        """
        Public method to send the raw input to the debugged program.
        
        @param s the raw input (string)
        """
        self.__sendCommand(s + '\n')
        self.emit(SIGNAL('clientRawInputSent'))
        
    def remoteClientVariables(self, scope, filter, framenr=0):
        """
        Public method to request the variables of the debugged program.
        
        @param scope the scope of the variables (0 = local, 1 = global)
        @param filter list of variable types to filter out (list of int)
        @param framenr framenumber of the variables to retrieve (int)
        """
        self.__sendCommand('%s%d, %d, %s\n' % \
            (RequestVariables, framenr, scope, str(filter)))
        
    def remoteClientVariable(self, scope, filter, var, framenr=0):
        """
        Public method to request the variables of the debugged program.
        
        @param scope the scope of the variables (0 = local, 1 = global)
        @param filter list of variable types to filter out (list of int)
        @param var list encoded name of variable to retrieve (string)
        @param framenr framenumber of the variables to retrieve (int)
        """
        self.__sendCommand('%s%s, %d, %d, %s\n' % \
            (RequestVariable, str(var), framenr, scope, str(filter)))
        
    def remoteClientSetFilter(self, scope, filter):
        """
        Public method to set a variables filter list.
        
        @param scope the scope of the variables (0 = local, 1 = global)
        @param filter regexp string for variable names to filter out (string)
        """
        self.__sendCommand('%s%d, "%s"\n' % (RequestSetFilter, scope, filter))
        
    def remoteEval(self, arg):
        """
        Public method to evaluate arg in the current context of the debugged program.
        
        @param arg the arguments to evaluate (string)
        """
        self.__sendCommand('%s%s\n' % (RequestEval,arg))
        
    def remoteExec(self, stmt):
        """
        Public method to execute stmt in the current context of the debugged program.
        
        @param stmt statement to execute (string)
        """
        self.__sendCommand('%s%s\n' % (RequestExec,stmt))
        
    def remoteBanner(self):
        """
        Public slot to get the banner info of the remote client.
        """
        self.__sendCommand(RequestBanner + '\n')
        
    def remoteCapabilities(self):
        """
        Public slot to get the debug clients capabilities.
        """
        self.__sendCommand(RequestCapabilities + '\n')
        
    def remoteCompletion(self, text):
        """
        Public slot to get the a list of possible commandline completions
        from the remote client.
        
        @param text the text to be completed (string or QString)
        """
        self.__sendCommand("%s%s\n" % (RequestCompletion, text))

    def remoteUTPrepare(self, fn, tn, tfn, cov, covname, coverase):
        """
        Public method to prepare a new unittest run.
        
        @param fn the filename to load (string)
        @param tn the testname to load (string)
        @param tfn the test function name to load tests from (string)
        @param cov flag indicating collection of coverage data is requested
        @param covname filename to be used to assemble the coverage caches
                filename
        @param coverase flag indicating erasure of coverage data is requested
        """
        # Restart the client if there is already a program loaded.
        try:
            self.__setClientType(ClientTypeAssociations[os.path.splitext(fn)[1]])
        except KeyError:
            self.__setClientType('Python')    # assume it is a Python file
        self.startClient(False)
        
        fn = self.translate(os.path.abspath(unicode(fn)), 0)
        self.__sendCommand('%s%s|%s|%s|%d|%s|%d\n' % \
            (RequestUTPrepare,
             unicode(fn), str(tn), str(tfn), cov, str(covname), coverase))
        self.debugging = False
        
    def remoteUTRun(self):
        """
        Public method to start a unittest run.
        """
        self.__sendCommand('%s\n' % RequestUTRun)
        
    def remoteUTStop(self):
        """
        public method to stop a unittest run.
        """
        self.__sendCommand('%s\n' % RequestUTStop)
        
    def __parseClientLine(self):
        """
        Private method to handle data from the client.
        """
        while self.qsock and self.qsock.canReadLine():
            qs = self.qsock.readLine(self.readLineBufferSize)
            if self.codec is not None:
                us = self.codec.fromUnicode(QString(qs))
            else:
                us = qs
            line = str(us)
            if line.endswith(EOT):
                line = line[:-len(EOT)]
                if not line:
                    continue

##            print "Server: ", line          ##debug
            
            eoc = line.find('<') + 1
            
            # Deal with case where user has written directly to stdout
            # or stderr, but not line terminated and we stepped over the
            # write call, in that case the >line< will not be the first
            # string read from the socket...
            boc = line.find('>')
            if boc > 0 and eoc > boc:
                self.emit(SIGNAL('clientOutput'), line[:boc])
                line = line[boc:]
                eoc = line.find('<') + 1
                boc = line.find('>')
                
            if boc >= 0 and eoc > boc:
                # Emit the signal corresponding to the response.
                resp = line[boc:eoc]

                if resp == ResponseLine:
                    stack = eval(line[eoc:-1])
                    for s in stack:
                        s[0] = self.translate(s[0], 1)
                    cf = stack[0]
                    if self.__autoContinue:
                        self.__autoContinue = False
                        QTimer.singleShot(0, self.remoteContinue)
                    else:
                        self.emit(SIGNAL('clientLine'), cf[0], int(cf[1]))
                        self.emit(SIGNAL('clientStack'), stack)
                    continue

                if resp == ResponseVariables:
                    self.emit(SIGNAL('clientVariables'), line[eoc:-1])
                    continue

                if resp == ResponseVariable:
                    self.emit(SIGNAL('clientVariable'), line[eoc:-1])
                    continue

                if resp == ResponseOK:
                    self.emit(SIGNAL('clientStatement'), 0)
                    continue

                if resp == ResponseContinue:
                    self.emit(SIGNAL('clientStatement'), 1)
                    continue

                if resp == ResponseException:
                    exc = line[eoc:-1]
                    exc = self.translate(exc, 1)
                    self.emit(SIGNAL('clientException'), exc)
                    continue

                if resp == ResponseSyntax:
                    exc = line[eoc:-1]
                    exc = self.translate(exc, 1)
                    self.emit(SIGNAL('clientSyntaxError'), exc)
                    continue
                    
                if resp == ResponseExit:
                    self.emit(SIGNAL('clientExit'), line[eoc:-1])
                    if self.passive:
                        self.__passiveShutDown()
                    if Preferences.getDebugger("AutomaticReset"):
                        self.startClient(False)
                    continue
                
                if resp == ResponseClearBreak:
                    fn, lineno = line[eoc:-1].split(',')
                    lineno = int(lineno)
                    fn = self.translate(fn, 1)
                    self.emit(SIGNAL('clientClearBreak'), fn, lineno)
                    continue
                    
                if resp == ResponseBPConditionError:
                    fn, lineno = line[eoc:-1].split(',')
                    lineno = int(lineno)
                    fn = self.translate(fn, 1)
                    self.emit(SIGNAL('clientBreakConditionError'), fn, lineno)
                    continue
                    
                if resp == ResponseClearWatch:
                    cond = line[eoc:-1]
                    self.emit(SIGNAL('clientClearWatch'), cond)
                    continue
                    
                if resp == ResponseWPConditionError:
                    cond = line[eoc:-1]
                    self.emit(SIGNAL('clientWatchConditionError'), cond)
                    continue
                    
                if resp == ResponseRaw:
                    prompt, echo = eval(line[eoc:-1])
                    self.emit(SIGNAL('clientRawInput'), prompt, echo)
                    continue
                    
                if resp == ResponseBanner:
                    version, platform, dbgclient = eval(line[eoc:-1])
                    self.emit(SIGNAL('clientBanner'), version, platform, dbgclient)
                    continue
                    
                if resp == ResponseCapabilities:
                    cap, clType = eval(line[eoc:-1])
                    self.emit(SIGNAL('clientCapabilities'), cap, clType)
                    continue
                    
                if resp == ResponseCompletion:
                    clstring, text = line[eoc:-1].split('||')
                    cl = eval(clstring)
                    self.emit(SIGNAL('clientCompletionList'), cl, text)
                    continue
                    
                if resp == PassiveStartup:
                    fn, exc = line[eoc:-1].split('|')
                    exc = int(exc)
                    fn = self.translate(fn, 1)
                    self.__passiveStartUp(fn, exc)
                    continue
                    
                if resp == ResponseUTPrepared:
                    res, exc_type, exc_value = eval(line[eoc:-1])
                    self.emit(SIGNAL('utPrepared'), res, exc_type, exc_value)
                    continue
                    
                if resp == ResponseUTStartTest:
                    testname, doc = eval(line[eoc:-1])
                    self.emit(SIGNAL('utStartTest'), testname, doc)
                    continue
                    
                if resp == ResponseUTStopTest:
                    self.emit(SIGNAL('utStopTest'))
                    continue
                    
                if resp == ResponseUTTestFailed:
                    testname, traceback = eval(line[eoc:-1])
                    self.emit(SIGNAL('utTestFailed'), testname, traceback)
                    continue
                    
                if resp == ResponseUTTestErrored:
                    testname, traceback = eval(line[eoc:-1])
                    self.emit(SIGNAL('utTestErrored'), testname, traceback)
                    continue
                    
                if resp == ResponseUTFinished:
                    self.emit(SIGNAL('utFinished'))
                    continue
                    
                if resp == ResponseCyclopsError:
                    fn, modfunc = eval(line[eoc:-1])
                    fn = self.translate(fn, 1)
                    self.emit(SIGNAL('cyclopsError'), fn, modfunc)
                    continue

            self.emit(SIGNAL('clientOutput'), line)

    def __passiveStartUp(self, fn, exc):
        """
        Private method to handle a passive debug connection.
        
        @param fn filename of the debugged script (string)
        @param exc flag to enable exception reporting of the IDE (boolean)
        """
        print unicode(self.trUtf8("Passive debug connection received"))
        self.passiveClientExited = False
        self.debugging = True
        self.__restoreBreakpoints()
        self.__restoreWatchpoints()
        self.emit(SIGNAL('passiveDebugStarted'), fn, exc)
        
    def __passiveShutDown(self):
        """
        Private method to shut down a passive debug connection.
        """
        self.passiveClientExited = True
        self.shutdownServer()
        print unicode(self.trUtf8("Passive debug connection closed"))
        
    def __setClientType(self, clType):
        """
        Private method to set the client type.
        
        @param clType type of client to be started (string, Python or Ruby)
        """
        if clType is not None and clType in ["Python", "Ruby"]:
            self.clientType = clType
            ok = Preferences.Prefs.settings.setValue('DebugClient/Type', 
                QVariant(self.clientType))
        
    def startClient(self, unplanned = True, clType = None, forProject = False):
        """
        Public method to start a debug client.
        
        @param unplanned flag indicating that the client has died (boolean)
        @param clType type of client to be started (string, Python or Ruby)
        @param forProject flag indicating a project related action (boolean)
        """
        if not self.passive or not self.passiveClientExited: 
            if self.qsock is not None:
                self.shutdownServer()
                self.emit(SIGNAL('clientGone'), unplanned & self.debugging)
            
        self.__setClientType(clType)
        self.startRemote(forProject)

    def __sendCommand(self,cmd):
        """
        Private method to send a single line command to the client.
        
        @param cmd command to send to the debug client (string)
        """
        if self.qsock is not None:
            self.qsock.write(str(cmd))
        else:
            self.queue.append(cmd)

    def __restoreBreakpoints(self):
        """
        Private method to restore the breakpoints after a restart.
        """
        if self.debugging:
            self.__addBreakPoints(QModelIndex(), 0, self.breakpointModel.rowCount()-1)
    
    def __restoreWatchpoints(self):
        """
        Private method to restore the watchpoints after a restart.
        """
        if self.debugging:
            self.__addWatchPoints(QModelIndex(), 0, self.watchpointModel.rowCount()-1)
    
    def getBreakPointModel(self):
        """
        Public slot to get a reference to the breakpoint model object.
        
        @return reference to the breakpoint model object (BreakPointModel)
        """
        return self.breakpointModel

    def getWatchPointModel(self):
        """
        Public slot to get a reference to the watchpoint model object.
        
        @return reference to the watchpoint model object (WatchPointModel)
        """
        return self.watchpointModel
