# -*- coding: utf-8 -*-

# Copyright (c) 2006 Detlev Offenbach <detlev@die-offenbachs.de>
#

"""
Module implementing a canvas class.
"""

from PyQt4.QtCore import *
from PyQt4.QtGui import *

class E4Canvas(QWidget):
    """
    Class implementing a canvas.
    """
    def __init__(self, parent, width_, height_, backColor = QColor(255, 255, 255)):
        """
        Constructor
        
        @param parent parent widget (QWidget)
        @param width_ initial width of the canvas (float)
        @param height_ initial height of the canvas (float)
        @param backColor background color of the canvas (QColor)
        """
        QWidget.__init__(self, parent)
        self.setObjectName("E4Canvas")
        
        self._items = []
        self._zoom = 1.0
        self._canvasSize = QSizeF(0.0, 0.0)
        
        palette = QPalette();
        palette.setColor(self.backgroundRole(), backColor);
        self.setPalette(palette);
        
        self.resize(width_, height_);
        self.setCanvasSize(width_, height_);
        self.update();
        
    def addItem(self, canvasItem):
        """
        Public method to add an item to the canvas.
        
        @param canvasItem item to be added to the canvas (E4CanvasItem)
        """
        self._items.append(canvasItem)
        
    def removeItem(self, canvasItem):
        """
        Public method to remove an item from the canvas.
        
        @param canvasItem item to be removed from the canvas (E4CanvasItem)
        """
        try:
            self._items.remove(canvasItem)
        except ValueError:
            pass    # there is no such item
        
    def getItems(self):
        """
        Public method to get all items of the canvas.
        
        @return reference to the list of items (list of E4CanvasItem)
        """
        return self._items
        
    def getSelectedItems(self):
        """
        Public method to retrieve all selected items.
        
        @return reference to the list of items (list of E4CanvasItem)
        """
        items = []
        for itm in self._items:
            itm.isSelected() and items.append(itm)
        return items
        
    def clear(self):
        """
        Public method to clear the canvas.
        
        <b>Note</b>: This does not delete the items.
        """
        self._items = []
        
    def paintEvent(self, evt):
        """
        Protected overridden method to paint the canvas.
        
        @param evt (QPaintEvent)
        """
        painter = QPainter(self)
        painter.setClipRect(evt.rect())
        
        painter.setRenderHint(QPainter.Antialiasing, True)
        painter.setRenderHint(QPainter.TextAntialiasing, True)
        painter.setRenderHint(QPainter.SmoothPixmapTransform, True)
        
        if self._zoom != 1.0:
            painter.scale(self._zoom, self._zoom)
        painter.fillRect(0, 0, 
                         int(self._canvasSize.width()), int(self._canvasSize.height()),
                         self.palette().color(self.backgroundRole()))
        for item in self._items:
            painter.save()
            item.drawShape(painter)
            painter.restore()
        painter.end()
        
    def sizeHint(self):
        """
        Public method to give a size hint.
        
        @return size hint (QSize)
        """
        return QSize(300, 200)
        
    def setCanvasSize(self, width, height):
        """
        Public method to set the canvas size.
        
        @param width new width of the canvas (float)
        @param height new height of the canvas (float)
        """
        self._canvasSize.setWidth(width)
        self._canvasSize.setHeight(height)
        
        self.resize(int(width * self._zoom),
                    int(height * self._zoom))
        self.update()
        
    def getCanvasSize(self):
        """
        Public method to get the canvas size.
        
        @return canvas size (QSizeF)
        """
        return self._canvasSize
        
    def getCanvasWidth(self):
        """
        Public method to get the canvas width.
        
        @return canvas width (float)
        """
        return self._canvasSize.width()
        
    def getCanvasHeight(self):
        """
        Public method to get the canvas height.
        
        @return height width (float)
        """
        return self._canvasSize.height()
        
    def resizeEvent(self, evt):
        """
        Protected overridden method to handle resize events.
        
        @param evt the resize event object (QResizeEvent)
        """
##        if self._zoom == 1.0:
        self._canvasSize.setWidth(evt.size().width() / self._zoom)
        self._canvasSize.setHeight(evt.size().height() / self._zoom)
        QWidget.resizeEvent(self, evt)
        
    def setZoom(self, zoom):
        """
        Public method to set the zoom level.
        
        @param zoom zoom level (float)
        """
        if zoom > 0.0:
            self._zoom = zoom
            self.resize(int(self._canvasSize.width() * self._zoom),
                        int(self._canvasSize.height() * self._zoom))
            self.update()
        
    def zoom(self):
        """
        Public method to get the zoom level.
        
        @return zoom level (float)
        """
        return self._zoom
        
    def collisions(self, pointOrRect):
        """
        Public method returning all items that collide with a point or rectangle.
        
        @param pointOrRect point or rectangle to chack against 
            (QPoint, QPointF, QRect or QRectF)
        @return list of references to items colliding with the given point or rectangle
        """
        collisionItems = []
        if isinstance(pointOrRect, QPoint) or isinstance(pointOrRect, QPointF):
            # it is a point
            if isinstance(pointOrRect, QPoint):
                pointf = QPointF(pointOrRect)
            else:
                pointf = pointOrRect
            for itm in self._items:
                if itm.boundingRect().contains(pointf):
                    collisionItems.append(itm)
        elif isinstance(pointOrRect, QRect) or isinstance(pointOrRect, QRectF):
            # it is a rectangle
            if isinstance(pointOrRect, QRect):
                rectf = QRectF(pointOrRect)
            else:
                rectf = pointOrRect
            for itm in self._items:
                if itm.boundingRect().intersects(rectf):
                    collisionItems.append(itm)
        return collisionItems
