# -*- coding: utf-8 -*-

# Copyright (c) 2004 - 2006 Detlev Offenbach <detlev@die-offenbachs.de>
#

"""
Module implementing a dialog showing an imports diagram of the application.
"""

import os

from PyQt4.QtCore import *
from PyQt4.QtGui import *

from KdeQt.KQProgressDialog import KQProgressDialog

from E4Canvas.E4Canvas import E4Canvas
from E4Canvas.E4CanvasText import E4CanvasText

from UMLCanvasView import UMLCanvasView
from PackageWidget import PackageWidget, PackageModel
from AssociationWidget import AssociationWidget, Imports
import GraphicsUtilities

from Utilities.ModuleParser import readModule
import Utilities

class ApplicationDiagram(QDialog):
    """
    Class implementing a dialog showing an imports diagram of the application.
    """
    allClasses = {}
    
    def __init__(self, project, parent = None, name = None,  noModules = 0):
        """
        Constructor
        
        @param project reference to the project object
        @param parent parent widget of the view (QWidget)
        @param name name of the view widget (QString or string)
        @keyparam noModules flag indicating, that no module names should be 
            shown (boolean)
        """
        QDialog.__init__(self, parent)

        if not name:
            self.setObjectName("ApplicationDiagram")
        else:
            self.setObjectName(name)

        self.UMLFormLayout = QVBoxLayout(self)
        self.UMLFormLayout.setMargin(6)
        self.UMLFormLayout.setSpacing(6)
        self.UMLFormLayout.setObjectName("UMLFormLayout")
        self.setLayout(self.UMLFormLayout)
        
        self.canvas = E4Canvas(None, 800, 600)
        self.umlCanvas = UMLCanvasView(self.canvas, self, "umlCanvas")
        self.UMLFormLayout.addWidget(self.umlCanvas)
        
        # generate the buttons
        buttonLayout = QHBoxLayout()
        buttonLayout.setMargin(0)
        buttonLayout.setSpacing(6)
        buttonLayout.setObjectName("layout1")
        
        leftSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        buttonLayout.addItem(leftSpacer)
        
        self.closeButton = QPushButton(self)
        self.closeButton.setObjectName("closeButton")
        self.closeButton.setDefault(True)
        self.closeButton.setText(self.trUtf8("&Close"))
        self.closeButton.setShortcut(self.trUtf8("Alt+C"))
        buttonLayout.addWidget(self.closeButton)
        self.connect(self.closeButton, SIGNAL("clicked()"), self, SLOT("close()"))
        
        rightSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        buttonLayout.addItem(rightSpacer)
        
        self.UMLFormLayout.addLayout(buttonLayout)
        
        self.project = project
        self.noModules = noModules
        
    def getDiagramName(self):
        """
        Method to retrieve a name for the diagram.
        
        @return name for the diagram
        """
        return self.project.ppath
        
    def __buildModulesDict(self):
        """
        Private method to build a dictionary of modules contained in the application.
        
        @return dictionary of modules contained in the application.
        """
        moduleDict = {}
        mods = self.project.pdata["SOURCES"]
        modules = []
        for module in mods:
            modules.append(Utilities.normabsjoinpath(self.project.ppath, module))
        tot = len(modules)
        try:
            prog = 0
            progress = KQProgressDialog(self.trUtf8("Parsing modules..."),
                QString(), 0, tot, self)
            progress.show()
            QApplication.processEvents()
            for module in modules:
                progress.setValue(prog)
                QApplication.processEvents()
                prog += 1
                if module.endswith("__init__.py"):
                    continue
                try: 
                    mod = readModule(module)
                except ImportError:
                    continue
                else:
                    name = mod.name
                    moduleDict[name] = mod
        finally:
            progress.setValue(tot)
        return moduleDict
        
    def __buildPackages(self):
        """
        Private method to build the modules shapes of the diagram.
        """
        project = os.path.splitdrive(self.project.ppath)[1].replace(os.sep, '.')[1:]
        packages = {}
        shapes = {}
        p = 10
        y = 10
        maxHeight = 0
        
        modules = self.__buildModulesDict()
        sortedkeys = modules.keys()
        sortedkeys.sort()
        
        # step 1: build a dictionary of packages
        for module in sortedkeys:
            l = module.split('.')
            package = '.'.join(l[:-1])
            if packages.has_key(package):
                packages[package][0].append(l[-1])
            else:
                packages[package] = ([l[-1]], [])
                
        # step 2: assign modules to dictionaries and update import relationship
        for module in sortedkeys:
            l = module.split('.')
            package = '.'.join(l[:-1])
            impLst = []
            for i in modules[module].imports:
                if modules.has_key(i):
                    impLst.append(i)
                else:
                    if i.find('.') == -1:
                        n = "%s.%s" % (modules[module].package, i)
                        if modules.has_key(n):
                            impLst.append(n)
                        else:
                            n = "%s.%s" % (project, i)
                            if modules.has_key(n):
                                impLst.append(n)
                            elif packages.has_key(n):
                                n = "%s.<<Dummy>>" % n
                                impLst.append(n)
                    else:
                        n = "%s.%s" % (project, i)
                        if modules.has_key(n):
                            impLst.append(n)
            for i in modules[module].from_imports.keys():
                if modules.has_key(i):
                    impLst.append(i)
                else:
                    if i.find('.') == -1:
                        n = "%s.%s" % (modules[module].package, i)
                        if modules.has_key(n):
                            impLst.append(n)
                        else:
                            n = "%s.%s" % (project, i)
                            if modules.has_key(n):
                                impLst.append(n)
                            elif packages.has_key(n):
                                n = "%s.<<Dummy>>" % n
                                impLst.append(n)
                    else:
                        n = "%s.%s" % (project, i)
                        if modules.has_key(n):
                            impLst.append(n)
            for imp in impLst:
                impPackage = '.'.join(imp.split('.')[:-1])
                if not impPackage in packages[package][1] and \
                   not impPackage == package:
                    packages[package][1].append(impPackage)
                    
        sortedkeys = packages.keys()
        sortedkeys.sort()
        for package in sortedkeys:
            if package:
                relPackage = package.replace(project, '')
                if relPackage and relPackage[0] == '.':
                    relPackage = relPackage[1:]
                else:
                    relPackage = unicode(self.trUtf8("<<Application>>"))
            else:
                relPackage = unicode(self.trUtf8("<<Others>>"))
            shape = self.__addPackage(relPackage, packages[package][0], 10.0, 10.0)
            shapes[package] = (shape, packages[package][1])
            pn = p + shape.width() + 10
            maxHeight = max(maxHeight, shape.height())
            if pn > self.canvas.getCanvasWidth():
                p = 10
                y += maxHeight + 10
                maxHeight = shape.height()
                shape.move(p, y)
                p += shape.width() + 10
            else:
                shape.move(p, y)
                p = pn
        
        self.__createAssociations(shapes)
        
    def __addPackage(self, name, modules, x, y):
        """
        Private method to add a package to the diagram.
        
        @param name package name to be shown (string)
        @param modules list of module names contained in the package
            (list of strings)
        @param x x-coordinate (float)
        @param y y-coordinate (float)
        """
        modules.sort()
        pm = PackageModel(name, modules)
        pw = PackageWidget(self.canvas, pm, x, y, noModules = self.noModules)
        pw.show()
        return pw
        
    def __createAssociations(self, shapes):
        """
        Private method to generate the associations between the package shapes.
        
        @param shapes list of shapes
        """
        for package in shapes.keys():
            for rel in shapes[package][1]:
                assoc = AssociationWidget(self.canvas, 
                        shapes[package][0], shapes[rel][0],
                        Imports)
                assoc.show()
        
    def show(self):
        """
        Overriden method to show the dialog.
        """
        self.__buildPackages()
        QDialog.show(self)
        
    def relayout(self):
        """
        Method to relayout the diagram.
        """
        self.allClasses.clear()
        self.__buildPackages()
