# -*- coding: utf-8 -*-

# Copyright (c) 2004 - 2006 Detlev Offenbach <detlev@die-offenbachs.de>
#

"""
Module implementing a canvas widget for an association between two widgets.
"""

from PyQt4.QtCore import *
from PyQt4.QtGui import *

from E4Canvas.E4CanvasArrow import E4CanvasArrow, NormalArrow, WideArrow

Normal = 0
Generalisation = 1
Imports = 2

NoRegion = 0
West = 1
North = 2
East = 3
South = 4
NorthWest = 5
NorthEast = 6
SouthEast = 7
SouthWest = 8
Center = 9

class AssociationWidget(E4CanvasArrow):
    """
    Class implementing a canvas widget for an association between two widgets.
    
    The association is drawn as an arrow starting at the first widget and
    ending at the second.
    """
    def __init__(self, canvas, widgetA, widgetB, type = Normal):
        """
        Constructor
        
        @param canvas canvas containing the association (QCanvas)
        @param widgetA first widget of the association
        @param widgetB second widget of the association
        @param type type of the association. This must be one of
            <ul>
            <li>Normal (default)</li>
            <li>Generalisation</li>
            <li>Imports</li>
            </ul>
        """
        if type == Normal:
            arrowType = NormalArrow
            arrowFilled = True
        elif type == Imports:
            arrowType = NormalArrow
            arrowFilled = True
        elif type == Generalisation:
            arrowType = WideArrow
            arrowFilled = False
        
        E4CanvasArrow.__init__(self, canvas, QPointF(0, 0), QPointF(100, 100), 
            arrowFilled, arrowType)
        
##        self.calculateEndingPoints = self.__calculateEndingPoints_center
        self.calculateEndingPoints = self.__calculateEndingPoints_rectangle
        
        self.widgetA = widgetA
        self.widgetB = widgetB
        self.assocType = type
        
        self.regionA = NoRegion
        self.regionB = NoRegion
        
        self.calculateEndingPoints()
        
        self.widgetA.addAssociation(self)
        self.widgetB.addAssociation(self)
        
    def __calculateEndingPoints_center(self):
        """
        Private method to calculate the ending points of the association widget.
        
        The ending points are calculated from the centers of the
        two associated widgets.
        """
        midA = QPointF(self.widgetA.x() + self.widgetA.width() / 2.0,
                       self.widgetA.y() + self.widgetA.height() / 2.0)
        midB = QPointF(self.widgetB.x() + self.widgetB.width() / 2.0,
                       self.widgetB.y() + self.widgetB.height() / 2.0)
        startP = self.__findRectIntersectionPoint(self.widgetA, midA, midB)
        endP = self.__findRectIntersectionPoint(self.widgetB, midB, midA)
        
        if startP.x() != -1 and startP.y() != -1 and \
           endP.x() != -1 and endP.y() != -1:
            self.setPoints(startP.x(), startP.y(), endP.x(), endP.y())
        
        self.update()
        
    def __calculateEndingPoints_rectangle(self):
        """
        Private method to calculate the ending points of the association widget.
        
        The ending points are calculated by the following method.
        
        For each Widget the diagram is divided in four Regions by its diagonals
        as indicated below
        <pre>
                   \  Region 2  /
                    \          /
                     |--------|
                     | \    / |
                     |  \  /  |
                     |   \/   |
            Region 1 |   /\   | Region 3
                     |  /  \  |
                     | /    \ |
                     |--------|
                    /          \
                   /  Region 4  \
        </pre>
        
        Each diagonal is defined by two corners of the bounding rectangle
        
        To calculate the start point  we have to find out in which
        region (defined by widgetA's diagonals) is widgetB's TopLeft corner 
        (lets call it region M) after that the start point will be
        the middle point of rectangle's side contained in region M.
        
        To calculate the end point we repeat the above  but in the opposite direction
        (from widgetB to widgetA)
        """
        if self.widgetA is None or self.widgetB is None:
            return
        
        xA = self.widgetA.x() + self.widgetA.width() / 2.0
        yA = self.widgetA.y() + self.widgetA.height() / 2.0
        xB = self.widgetB.x() + self.widgetB.width() / 2.0
        yB = self.widgetB.y() + self.widgetB.height() / 2.0
        
        # find widgetA region
        rc = QRectF(xA, yA, self.widgetA.width(), self.widgetA.height())
        oldRegionA = self.regionA
        self.regionA = self.__findPointRegion(rc, xB, yB)
        # move some regions to the standard ones
        if self.regionA == NorthWest:
            self.regionA = North
        elif self.regionA == NorthEast:
            self.regionA = East
        elif self.regionA == SouthEast:
            self.regionA = South
        elif self.regionA == SouthWest:
            self.regionA = West
        elif self.regionA == Center:
            self.regionA = West
        
        self.__updateEndPoint(self.regionA, True)
        
        # now do the same for widgetB
        rc = QRectF(xB, yB, self.widgetB.width(), self.widgetB.height())
        oldRegionB = self.regionB
        self.regionB = self.__findPointRegion(rc, xA, yA)
        # move some regions to the standard ones
        if self.regionB == NorthWest:
            self.regionB = North
        elif self.regionB == NorthEast:
            self.regionB = East
        elif self.regionB == SouthEast:
            self.regionB = South
        elif self.regionB == SouthWest:
            self.regionB = West
        elif self.regionB == Center:
            self.regionB = West
        
        self.__updateEndPoint(self.regionB, False)
        
        self.update()
        
    def __findPointRegion(self, rect, posX, posY):
        """
        Private method to find out, which region of rectangle rect contains the point 
        (PosX, PosY) and returns the region number.
        
        @param rect rectangle to calculate the region for (QRectF)
        @param posX x position of point (float)
        @param posY y position of point (float)
        @return the calculated region number<br />
            West = Region 1<br />
            North = Region 2<br />
            East = Region 3<br />
            South = Region 4<br />
            NorthWest = On diagonal 2 between Region 1 and 2<br />
            NorthEast = On diagonal 1 between Region 2 and 3<br />
            SouthEast = On diagonal 2 between Region 3 and 4<br />
            SouthWest = On diagonal 1 between Region4 and 1<br />
            Center = On diagonal 1 and On diagonal 2 (the center)<br />
        """
        w = rect.width()
        h = rect.height()
        x = rect.x()
        y = rect.y()
        slope2 = w / h
        slope1 = -slope2
        b1 = x + w / 2.0 - y * slope1
        b2 = x + w / 2.0 - y * slope2
        
        eval1 = slope1 * posY + b1
        eval2 = slope2 * posY + b2
        
        result = NoRegion
        
        # inside region 1
        if eval1 > posX and eval2 > posX:
            result = West
        
        #inside region 2
        elif eval1 > posX and eval2 < posX:
            result = North
        
        # inside region 3
        elif eval1 < posX and eval2 < posX:
            result = East
        
        # inside region 4
        elif eval1 < posX and eval2 > posX:
            result = South
        
        # inside region 5
        elif eval1 == posX and eval2 < posX:
            result = NorthWest
        
        # inside region 6
        elif eval1 < posX and eval2 == posX:
            result = NorthEast
        
        # inside region 7
        elif eval1 == posX and eval2 > posX:
            result = SouthEast
        
        # inside region 8
        elif eval1 > posX and eval2 == posX:
            result = SouthWest
        
        # inside region 9
        elif eval1 == posX and eval2 == posX:
            result = Center
        
        return result
        
    def __updateEndPoint(self, region, isWidgetA):
        """
        Private method to update an endpoint.
        
        @param region the region for the endpoint (integer)
        @param isWidgetA flag indicating update for widgetA is done (boolean)
        """
        if region == NoRegion:
            return
        
        if isWidgetA:
            widget = self.widgetA
        else:
            widget = self.widgetB
        x = widget.x()
        y = widget.y()
        ww = widget.width()
        wh = widget.height()
        ch = wh / 2.0
        cw = ww / 2.0
        
        if region == West:
            px = x
            py = y + ch
        elif region == North:
            px = x + cw
            py = y
        elif region == East:
            px = x + ww
            py = y + ch
        elif region == South:
            px = x + cw
            py = y + wh
        elif region == Center:
            px = x + cw
            py = y + wh
        
        if isWidgetA:
            self.setStartPoint(px, py)
        else:
            self.setEndPoint(px, py)
        
    def __findRectIntersectionPoint(self, widget, p1, p2):
        """
        Private method to find the intersetion point of a line with a rectangle.
        
        @param widget widget to check against
        @param p1 first point of the line (QPointF)
        @param p2 second point of the line (QPointF)
        @return the intersection point (QPointF)
        """
        result = QPointF(-1.0, -1.0)
        old_region = -1
        if widget is None:
            return result
        
        X = -1.0
        Y = -1.0
        if widget == self.widgetA:
            X = self.widgetB.x()
            Y = self.widgetB.y()
        else:
            X = self.widgetA.x()
            Y = self.widgetA.y()
            
        rc = QRectF(widget.x(), widget.y(), widget.width(), widget.height())
        region = self.__findPointRegion(rc, X, Y)
        if region == old_region:
            return QPointF(-1, -1)
        
        if region == West:
            result = self.__findIntersection(rc.topLeft(), rc.bottomLeft(), p1, p2)
        elif region == North:
            result = self.__findIntersection(rc.topLeft(), rc.topRight(), p1, p2)
        elif region == East:
            result = self.__findIntersection(rc.topRight(), rc.bottomRight(), p1, p2)
        elif region == South:
            result = self.__findIntersection(rc.bottomLeft(), rc.bottomRight(), p1, p2)
        elif region == NorthWest:
            result = rc.topLeft()
        elif region == NorthEast:
            result = rc.topRight()
        elif region == SouthEast:
            result = rc.bottomRight()
        elif region == SouthWest or region == Center:
            result = rc.bottomLeft()
        
        return result
        
    def __findIntersection(self, p1, p2, p3, p4):
        """
        Method to calculate the intersection point of two lines.
        
        The first line is determined by the points p1 and p2, the second
        line by p3 and p4. If the intersection point is not contained in
        the segment p1p2, then it returns (-1.0, -1.0).
        
        For the function's internal calculations remember:<br />
        QT coordinates start with the point (0,0) as the topleft corner
        and x-values increase from left to right and y-values increase
        from top to bottom; it means the visible area is quadrant I in
        the regular XY coordinate system
        
        <pre>
            Quadrant II     |   Quadrant I
           -----------------|-----------------
            Quadrant III    |   Quadrant IV
        </pre>
        
        In order for the linear function calculations to work in this method
        we must switch x and y values (x values become y values and viceversa)
        
        @param p1 first point of first line (QPointF)
        @param p2 second point of first line (QPointF)
        @param p3 first point of second line (QPointF)
        @param p4 second point of second line (QPointF)
        @return the intersection point (QPointF)
        """
        x1 = p1.y()
        y1 = p1.x()
        x2 = p2.y()
        y2 = p2.x()
        x3 = p3.y()
        y3 = p3.x()
        x4 = p4.y()
        y4 = p4.x()
        
        # line 1 is the line between (x1, y1) and (x2, y2)
        # line 2 is the line between (x3, y3) and (x4, y4)
        no_line1 = True    # it is false, if line 1 is a linear function
        no_line2 = True    # it is false, if line 2 is a linear function
        slope1 = 0.0
        slope2 = 0.0
        b1 = 0.0
        b2 = 0.0
        
        if x2 != x1:
            slope1 = (y2 - y1) / (x2 - x1)
            b1 = y1 - slope1 * x1
            no_line1 = False
        if x4 != x3:
            slope2 = (y4 - y3) / (x4 - x3)
            b2 = y3 - slope2 * x3
            no_line2 = False
        
        pt = QPointF()
        # if either line is not a function
        if no_line1 and no_line2:
            # if the lines are not the same one
            if x1 != x3:
                return QPointF(-1.0, -1.0)
            # if the lines are the same ones
            if y3 <= y4:
                if y3 <= y1 and y1 <= y4:
                    return QPointF(y1, x1)
                else:
                    return QPointF(y2, x2)
            else:
                if y4 <= y1 and y1 <= y3:
                    return QPointF(y1, x1)
                else:
                    return QPointF(y2, x2)
        elif no_line1:
            pt.setX(slope2 * x1 + b2)
            pt.setY(x1)
            if y1 >= y2:
                if not (y2 <= pt.x() and pt.x() <= y1):
                    pt.setX(-1.0)
                    pt.setY(-1.0)
            else:
                if not (y1 <= pt.x() and pt.x() <= y2):
                    pt.setX(-1.0)
                    pt.setY(-1.0)
            return pt
        elif no_line2:
            pt.setX(slope1 * x3 + b1)
            pt.setY(x3)
            if y3 >= y4:
                if not (y4 <= pt.x() and pt.x() <= y3):
                    pt.setX(-1.0)
                    pt.setY(-1.0)
            else:
                if not (y3 <= pt.x() and pt.x() <= y4):
                    pt.setX(-1.0)
                    pt.setY(-1.0)
            return pt
        
        if slope1 == slope2:
            pt.setX(-1.0)
            pt.setY(-1.0)
            return pt
        
        pt.setY((b2 - b1) / (slope1 - slope2))
        pt.setX(slope1 * pt.y() + b1)
        # the intersection point must be inside the segment (x1, y1) (x2, y2)
        if x2 >= x1 and y2 >= y1:
            if not ((x1 <= pt.y() and pt.y() <= x2) and (y1 <= pt.x() and pt.x() <= y2)):
                pt.setX(-1.0)
                pt.setY(-1.0)
        elif x2 < x1 and y2 >= y1:
            if not ((x2 <= pt.y() and pt.y() <= x1) and (y1 <= pt.x() and pt.x() <= y2)):
                pt.setX(-1.0)
                pt.setY(-1.0)
        elif x2 >= x1 and y2 < y1:
            if not ((x1 <= pt.y() and pt.y() <= x2) and (y2 <= pt.x() and pt.x() <= y1)):
                pt.setX(-1.0)
                pt.setY(-1.0)
        else:
            if not ((x2 <= pt.y() and pt.y() <= x1) and (y2 <= pt.x() and pt.x() <= y1)):
                pt.setX(-1.0)
                pt.setY(-1.0)
        
        return pt
        
    def widgetMoved(self):
        """
        Public method to recalculate the association after a widget was moved.
        """
        self.calculateEndingPoints()
        
    def unassociate(self):
        """
        Public method to unassociate from the widgets.
        """
        self.widgetA.removeAssociation(self)
        self.widgetB.removeAssociation(self)
