# -*- coding: utf-8 -*-

# Copyright (c) 2004 - 2006 Detlev Offenbach <detlev@die-offenbachs.de>
#

"""
Module implementing a dialog showing an imports diagram of a package.
"""

import glob
import os
import sys

from PyQt4.QtCore import *
from PyQt4.QtGui import *

from KdeQt.KQProgressDialog import KQProgressDialog

from E4Canvas.E4Canvas import E4Canvas
from E4Canvas.E4CanvasText import E4CanvasText

from UMLCanvasView import UMLCanvasView
from ModuleWidget import ModuleWidget, ModuleModel
from AssociationWidget import AssociationWidget, Imports
import GraphicsUtilities

from Utilities.ModuleParser import readModule
import Utilities

class ImportsDiagram(QDialog):
    """
    Class implementing a dialog showing an imports diagram of a package.
    
    Note: Only package internal imports are show in order to maintain
    some readability.
    """
    allClasses = {}
    
    def __init__(self, package, parent = None, name = None):
        """
        Constructor
        
        @param package name of a python package to show the import 
            relationships (string)
        @param parent parent widget of the view (QWidget)
        @param name name of the view widget (QString or string)
        """
        QDialog.__init__(self, parent)

        if not name:
            self.setObjectName("ImportsDiagram")
        else:
            self.setObjectName(name)

        self.UMLFormLayout = QVBoxLayout(self)
        self.UMLFormLayout.setMargin(6)
        self.UMLFormLayout.setSpacing(6)
        self.UMLFormLayout.setObjectName("UMLFormLayout")
        self.setLayout(self.UMLFormLayout)
        
        self.canvas = E4Canvas(None, 800, 600)
        self.umlCanvas = UMLCanvasView(self.canvas, self, "umlCanvas")
        self.UMLFormLayout.addWidget(self.umlCanvas)
        
        # generate the buttons
        buttonLayout = QHBoxLayout()
        buttonLayout.setMargin(0)
        buttonLayout.setSpacing(6)
        buttonLayout.setObjectName("layout1")
        
        leftSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        buttonLayout.addItem(leftSpacer)
        
        self.closeButton = QPushButton(self)
        self.closeButton.setObjectName("closeButton")
        self.closeButton.setDefault(True)
        self.closeButton.setText(self.trUtf8("&Close"))
        self.closeButton.setShortcut(self.trUtf8("Alt+C"))
        buttonLayout.addWidget(self.closeButton)
        self.connect(self.closeButton, SIGNAL("clicked()"), self, SLOT("close()"))
        
        rightSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        buttonLayout.addItem(rightSpacer)
        
        self.UMLFormLayout.addLayout(buttonLayout)
        
        self.packagePath = Utilities.normabspath(package)
        self.package = self.packagePath.replace(os.sep, '.')[1:]
        
    def getDiagramName(self):
        """
        Method to retrieve a name for the diagram.
        
        @return name for the diagram
        """
        return self.packagePath
        
    def __buildModulesDict(self):
        """
        Private method to build a dictionary of modules contained in the package.
        
        @return dictionary of modules contained in the package.
        """
        moduleDict = {}
        modules = glob.glob(Utilities.normjoinpath(self.packagePath,'*.py')) + \
                  glob.glob(Utilities.normjoinpath(self.packagePath,'*.ptl'))
        tot = len(modules)
        try:
            prog = 0
            progress = KQProgressDialog(self.trUtf8("Parsing modules..."),
                QString(), 0, tot, self)
            progress.show()
            QApplication.processEvents()
            for module in modules:
                progress.setValue(prog)
                QApplication.processEvents()
                prog = prog + 1
##                if module.endswith("__init__.py"):
##                    continue
                try: 
                    mod = readModule(module)
                except ImportError:
                    continue
                else:
                    name = mod.name
                    if name.startswith(self.package):
                        name = name[len(self.package) + 1:]
                    moduleDict[name] = mod
        finally:
            progress.setValue(tot)
        return moduleDict
        
    def __buildImports(self):
        """
        Private method to build the modules shapes of the diagram.
        """
        initlist = glob.glob(os.path.join(self.packagePath, '__init__.*'))
        if len(initlist) == 0:
            ct = E4CanvasText(self.canvas, QPointF(10.0, 10.0),
                self.trUtf8("The directory '%1' is not a Python package.")\
                    .arg(self.package))
            ct.show()
            return
        
        shapes = {}
        p = 10
        y = 10
        maxHeight = 0
        
        modules = self.__buildModulesDict()
        sortedkeys = modules.keys()
        sortedkeys.sort()
        for module in sortedkeys:
            impLst = []
            for i in modules[module].imports:
                if i.startswith(self.package):
                    n = i[len(self.package) + 1:]
                else:
                    n = i
                if modules.has_key(i):
                    impLst.append(n)
            for i in modules[module].from_imports.keys():
                if i.startswith(self.package):
                    n = i[len(self.package) + 1:]
                else:
                    n = i
                if modules.has_key(i):
                    impLst.append(n)
            classNames = []
            for cls in modules[module].classes.keys():
                className = modules[module].classes[cls].name
                if className not in classNames:
                    classNames.append(className)
##            shape = self.__addModule(module, classNames, p, y)
            shape = self.__addModule(module, classNames, 10.0, 10.0)
            shapes[module] = (shape, impLst)
            pn = p + shape.width() + 10
            maxHeight = max(maxHeight, shape.height())
            if pn > self.canvas.getCanvasWidth():
                p = 10
                y += maxHeight + 10
                maxHeight = shape.height()
                shape.move(p, y)
                p += shape.width() + 10
            else:
                shape.move(p, y)
                p = pn
        
        self.__createAssociations(shapes)
        
    def __addModule(self, name, classes, x, y):
        """
        Private method to add a module to the diagram.
        
        @param name module name to be shown (string)
        @param classes list of class names contained in the module
            (list of strings)
        @param x x-coordinate (float)
        @param y y-coordinate (float)
        """
        classes.sort()
        impM = ModuleModel(name, classes)
        impW = ModuleWidget(self.canvas, impM, x, y)
        impW.show()
        return impW
        
    def __createAssociations(self, shapes):
        """
        Private method to generate the associations between the module shapes.
        
        @param shapes list of shapes
        """
        for module in shapes.keys():
            for rel in shapes[module][1]:
                assoc = AssociationWidget(self.canvas, 
                        shapes[module][0], shapes[rel][0],
                        Imports)
                assoc.show()
        
    def show(self):
        """
        Overriden method to show the dialog.
        """
        self.__buildImports()
        QDialog.show(self)
        
    def relayout(self):
        """
        Method to relayout the diagram.
        """
        self.allClasses.clear()
        self.__buildImports()
